/***********************************************************************
 * File: EUC_code_counter.c
 * 	(Extracted from
 *	   "$Id: EUC_code_counter.c,v 1.1 2007/10/28 06:03:19 Administrator Exp Administrator $")
 *
 * Abstracts:
 *  This C source implements a character counter for the EUC-JP encoded
 *  strings, and getting sub-string of the EUC-JP encode string.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/


/* include this library definitions */
#include "EUC_code_counter.h"


/* definitions of EUC-JP encoded character sizes */
#define	ASCII_SZ		1
#define JIS_X_0208_1983_SZ	2
#define JIS_X_0201_1976_SZ	2
#define JIS_X_0212_1990_SZ	3
#define	UNKNOWN_CODE_SZ		1


/* definitions of EUC-JP encoded character classes */
#define IS_ASCII(STR)		((unsigned)(*(STR)) <= 0x7f)
#define IS_ASCII_CONTROLS(STR)	(((unsigned)(*(STR)) <= 0x1f) \
				|| (unsigned)(*(STR)) == 0x7f)

#define IS_JIS_X(STR) ((0xa1 <= (unsigned)(*(STR)))		\
			&& (unsigned)(*(STR)) <= 0xfe)
#define IS_JIS_X_0208_1983(STR)	(				\
			(IS_JIS_X((STR)))			\
			&& IS_JIS_X(((STR) + 1))		\
		) /* Level 1 and 2 Japanese Kanji */

#define IS_JIS_X_0201_1976(STR) (				\
			(0x8e == (unsigned)(*(STR)))		\
			&& IS_JIS_X(((STR) + 1))		\
		) /* Japanese Hankaku-Kana */

#define IS_JIS_X_0212_1990(STR)	(				\
			(0x8f == (unsigned)(*(STR)))  		\
			&& IS_JIS_X_0208_1983(((STR) + 1))	\
		) /* Level 3 (Gaiji) Japanese Kanji */


/* count up the EUC-JP characters in a string */
size_t count_euc_code_characters(unsigned char *str
			, size_t	str_sz
			, EUC_code_character_class_counts *counts
			)
{
	size_t	charsz, sz = 0;

	/* initialize errno */
	errno	= 0;

	/* test: illegal arguments? */
	if (str == NULL || counts == NULL) {
		errno	= EINVAL;
		return	0; /* FAILED: illegal arguments */
	}

	/* initialize the output */
	errno	= 0;
	counts->count_of_ASCII	= counts->count_of_JIS_X_0208_1983
				= counts->count_of_JIS_X_0201_1976
				= counts->count_of_JIS_X_0212_1990
				= counts->count_of_UNKNOWN
				= counts->count_of_new_lines
				= counts->length_of_bytes
				= 0;

	/* count up the counters... */
	while ((sz < str_sz) && (char)(*str) != '\0') {
		if (IS_ASCII(str)) {
			++(counts->count_of_ASCII);
			if (IS_ASCII_CONTROLS(str)) {
				++(counts->count_of_controls);
				if (*str == (unsigned char)'\n') {
					++(counts->count_of_new_lines);
				}
			}
			charsz	= ASCII_SZ;
		}
		else if ((sz + JIS_X_0208_1983_SZ - 1 < str_sz)
		&& IS_JIS_X_0208_1983(str)) {
			++(counts->count_of_JIS_X_0208_1983);
			charsz	= JIS_X_0208_1983_SZ;
		}
		else if ((sz + JIS_X_0201_1976_SZ - 1 < str_sz)
		&& IS_JIS_X_0201_1976(str)) {
			++(counts->count_of_JIS_X_0201_1976);
			charsz	= JIS_X_0201_1976_SZ;
		}
		else if ((sz + JIS_X_0212_1990_SZ - 1 < str_sz)
		&& IS_JIS_X_0212_1990(str)) {
			++(counts->count_of_JIS_X_0212_1990);
			charsz	= JIS_X_0212_1990_SZ;
		}
		else {	/* Unknown character in EUC-JP encoding */
			++(counts->count_of_UNKNOWN);
			charsz	= UNKNOWN_CODE_SZ;
		}
		counts->length_of_bytes	+= charsz;
		sz	+= charsz;
		str	+= charsz;
	}

	if (counts->count_of_UNKNOWN > 0) {
		errno	= EINVAL; /* str isn't encodes in EUC-JP */
	}

	/* returns total counts of characters */
	return	counts->count_of_ASCII
			+ counts->count_of_JIS_X_0208_1983
			+ counts->count_of_JIS_X_0201_1976
			+ counts->count_of_JIS_X_0212_1990
			+ counts->count_of_UNKNOWN
			;
}


/* get width of EUC-JP character */
static size_t get_width_as_single_character(unsigned char	*str
					, size_t		sz
					, void			*usrdata
					)
{
	/* get the width as single-character */
	if ((sz == ASCII_SZ)
	|| ((unsigned)(*str) == 0x8e && sz == JIS_X_0201_1976_SZ)) {
		return	1;	/* ASCII or JIS X 0201 1976 */
	}
	else if ((sz == JIS_X_0208_1983_SZ)
	|| sz == JIS_X_0212_1990_SZ) {
		return	2;	/* JIS X 0208 1983 or JIS X 0212 1990 */
	}

	/* FAILED: illegal arguments */
	errno	= EINVAL;
	return	0;
}


/* get the size of a character */
static unsigned char *get_character_size(unsigned char *str
					, size_t *charsz
					)
{
	if (IS_ASCII(str)) {
		*charsz	= ASCII_SZ;
	}
	else if (IS_JIS_X_0208_1983(str)) {
		*charsz	= JIS_X_0208_1983_SZ;
	}
	else if (IS_JIS_X_0201_1976(str)) {
		*charsz	= JIS_X_0201_1976_SZ;
	}
	else if (IS_JIS_X_0212_1990(str)) {
		*charsz	= JIS_X_0212_1990_SZ;
	}
	else {	/* Unknown character in EUC-JP */
		return	NULL;
	}

	return	str;
}


/* get address of the sub-string */
static unsigned char *get_address_of_sub_string(unsigned char	*str
						, size_t *sz
						, size_t *nth_char
						)
{
	size_t	charsz = 0;

	while ((*str != (unsigned char)'\0') && *nth_char > 0) {
		if (get_character_size(str, &charsz) == NULL) {
			/* Unknown character in EUC-JP encoding */
			*sz	= 0;
			return	NULL;	/* FAILED: illegal args */
		}

		*sz	+= charsz;
		str	+= charsz;
		--(*nth_char);
	}

	return	str;	/* Normal End */
}

/*
 * get sub-string address of a EUC-JP string,
 * and output size of sub-string
 */
char *substr_of_euc_code_string(unsigned char	*str
				, size_t	str_sz
				, size_t	nth_char
				, size_t	n_chars
				, size_t	w_chars
				, char		*delimits
				, size_t	(*metrics)(
						  unsigned char *str
						, size_t	sz
						, void		*usrdata
						)
				, void		*usrdata
				, size_t	*size
				, size_t	*width
				)
{
	char	*result;
	size_t	charsz = 0, sz = 0;

	/* initialize errno */
	errno	= 0;

	/* test: illegal arguments? */
	if ((((n_chars < 1 && w_chars < 1)
	|| str == NULL) || size == NULL) || width == NULL) {
		errno	= EINVAL;
		return	NULL; /* FAILED: illegal arguments */
	}

	/* initialize the outputs */
	*size = 0; *width = 0;

	/* get address of the sub-string... */
	if ((str = get_address_of_sub_string(str
					, &sz
					, &nth_char
					)) == NULL) {
		errno	= EINVAL;
		return	NULL; /* FAILED: illegal arguments */
	}

	/* test: 'nth_char' arg is illegal? */
	if (((nth_char > 0) || sz >= str_sz)
	|| *str == (unsigned char)'\0') {
		errno	= EDOM;
		return	NULL; /* FAILED: out of string */
	}

	/* set the result */
	result = (char *)str;

	/* get the size of sub-string... */
	if (n_chars > 0) {
		while (((sz < str_sz) && *str != (unsigned char)'\0')
		&& n_chars > 0) {
			if (get_character_size(str, &charsz) == NULL) {
				/* Unknown character in EUC-JP */
				*size	= 0;
				errno	= EINVAL;
				return	NULL;  /* FAILED:
						* illegal arguments
						*/
			}

			*size	+= charsz;
			str	+= charsz;
			sz	+= charsz;
			*width	+= charsz;
			--n_chars;
		}
	}
	else {	/* width of sub-string equals or less than 'w_chars' */
		size_t	w = 0;

		/* test: metrics is default? */
		if (metrics == NULL) {
			/* set default metrics */
			metrics	= get_width_as_single_character;
		}

		while (((sz < str_sz)
		&& *str != (unsigned char)'\0')
		&& (delimits == NULL
			|| strchr(delimits, (int)(*str)) == NULL)
		) {
			if (get_character_size(str, &charsz) == NULL) {
				/* Unknown character in EUC-JP */
				*size	= 0;
				errno	= EINVAL;
				return	NULL;  /* FAILED:
						* illegal arguments
						*/
			}

			/* get width of this character */
			if ((w = (*metrics)(str
					, charsz
					, usrdata
					)) == 0UL) {
				*size	= 0;
				errno	= EINVAL;
				return	NULL;  /* FAILED:
						* illegal arguments
						*/
			}

			/* test: width over run? */
			if (*width + w > w_chars) {
				break; /* enough characters */
			}

			*width	+= w;
			*size += charsz;
			str += charsz;
			sz	+= charsz;
		}
	}

	return	result; /* Normal End */
}

/* trimming the EUC-JP wide character spaces and ASCII spaces */
char *trim_EUCJP_spaces(Trim_EUCJP_spaces_direction_t direct
			, char *str
			)
{
	unsigned char	EUCJP_space[] = { (unsigned char)0xa1
					, (unsigned char)0xa1
					, '\0'
					}; /* EUC-JP wide-char space */
	char		*sp;

	/* test: illegal arguments? */
	if ((((direct & ~(Trim_EUCJP_spaces_LEFT
				| Trim_EUCJP_spaces_RIGHT
				| Trim_ASCII_all_space_codes)) != 0)
	|| (direct & ~Trim_ASCII_all_space_codes) == 0)
	|| str == NULL) {
		errno = EINVAL; return NULL; /* ERROR: invalid args */
	}

	/* test: we must trim the spaces at the head of string? */
	if ((direct & Trim_EUCJP_spaces_RIGHT) != 0) {
		/* trim the head spaces... */
		for (sp = str + strlen(str); sp-- > str; ) {
			if (((sp - str) >= 1)
			&& strncmp(sp - 1
				, (char *)EUCJP_space
				, sizeof(EUCJP_space) - 1
				) == 0) {
				sp	-= (sizeof(EUCJP_space) - 2);
			}
			else if ((((direct & Trim_ASCII_all_space_codes)
					!= 0) && !isspace(*sp))
			|| *sp != ' ') {
				break;
			}
		}
		*(sp + 1)	= '\0'; /* trim the tail spaces */
	}

	/* test: we must trim the spaces at the tail of string? */
	if ((direct & Trim_EUCJP_spaces_LEFT) != 0) {
		/* find the non-space code... */
		for (sp = str; *sp != '\0'; ) {
			if (strncmp(sp
				, (char *)EUCJP_space
				, sizeof(EUCJP_space) - 1
				) == 0) {
				sp	+= (sizeof(EUCJP_space) - 1);
			}
			else if ((((direct & Trim_ASCII_all_space_codes)
					!= 0) && isspace(*sp))
				|| *sp == ' ') {
				++sp;
			}
			else {
				break;
			}
		}
		/* test: found the spaces? */
		if (sp != str) {
			strcpy(str, sp); /* trim the head spaces */
		}
	}

	return	str; /* result of trimming */
}

/* copy a string that encoded EUC-JP */
char *strcpy_EUCJP(unsigned char *src, char *dst, size_t dst_sz)
{
	char		*sv_dst = dst;
	size_t		src_sz, size, width, w_chars;

	/* test: illegal arguments? */
	if (((src == NULL) || dst == NULL) || dst_sz == 0) {
		errno = EINVAL; return NULL; /* ERROR: invalid args */
	}

	/* processing the src... */
	src_sz	= strlen((char *)src); /* get the length of src */
	--dst_sz; /* increment dst_sz for nil-code */
	while ((src_sz > 0) && dst_sz > 0) {
		/* get the width of destiny */
		if (dst_sz >= (2 * JIS_X_0212_1990_SZ)) {
			/* require 1/3 characters of destiny size */
			w_chars	= dst_sz / JIS_X_0212_1990_SZ;
		}
		else if (dst_sz >= JIS_X_0212_1990_SZ) {
			/* destiny size 3 to 5:
			 * => require 2 character width
			 */
			w_chars	= 2;
		}
		else {	/* destiny size 1 or 2:
			 * => require destiny size
			 */
			w_chars	= dst_sz;
		}

		/* try to get the size of transfer characters... */
		size	= 0; /* initialize the size (output) */
		width	= 0; /* initialize the width (output) */
		if (substr_of_euc_code_string(src
					, src_sz
					, 0
					, 0
					, w_chars
					, NULL
					, NULL
					, NULL
					, &size
					, &width
					) == NULL) {
			/* ERROR: illegal encoding */
			return	NULL;
		}
			
		if (size > dst_sz) {
			/* test: the destiny size is less than 6 bytes,
			 *   and the require size (w_chars) is 2 bytes?
			 */
			if ((dst_sz < (2 * JIS_X_0212_1990_SZ))
			&& w_chars == 2) {
				/* try to get the 1 character width...
				 */
				w_chars	= 1;
				if (substr_of_euc_code_string(src
							, src_sz
							, 0
							, 0
							, w_chars
							, NULL
							, NULL
							, NULL
							, &size
							, &width
							) == NULL) {
					/* ERROR: illegal encoding */
					return	NULL;
				}
			}
			else {
				break; /* normal end */
			}
		}

		/* test: transfer the all acceptable characters? */
		if (size == 0) {
			break; /* normal end */
		}

		/* transfer the characters */
		memcpy(dst, src, size);
		dst += size; dst_sz -= size;
		src += size; src_sz -= size;
	}

	*dst	= '\0'; /* terminate the destiny string */

	return	sv_dst; /* the address of destiny string */
}
